/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
public struct Point {
	public double x;
	public double y;

	public static double hypot (Point p0, Point p1) {
		return Math.hypot (p1.y - p0.y, p1.x - p0.x);
	}

	public Point rotate (double angle) {
		return Point () {
			x = x * Math.cos (angle) - y * Math.sin (angle),
			y = x * Math.sin (angle) + y * Math.cos (angle)};
	}

	public Point scale (double sx, double sy) {
		return Point () {x = x * sx, y = y * sy};
	}

	public Point move (Point new_origin) {
		return Point () {x = x + new_origin.x, y = y + new_origin.y};
	}

	public Point move_xy (double x, double y) {
		return Point () {x = this.x + x, y = this.y + y};
	}
}

public enum Role {
	/**
	 * Seed replicated starting at depth 0
	 */
	DUAL = 0,
	/**
	 * Seed replicated starting at depth 1
	 */
	MOTIF = 1,
	/**
	 * Seed replicated at depth 0, but disappears at depth 1
	 */
	BASE = 2,
}

public enum RenderStyle {
	THIN = 0,
	THICK = 1,
	HOLLOW = 2,
	FILLED = 3,
	CLOSED = 4,
	NONE = 5,
}

public class Projection {
	weak Seed seed;
	Projection? parent;

	// Precalculated affine projection. Precalculation is done in constructor.
	//  [a b 0]
	//  [c d 0]
	//  [e f 1]
	double a;
	double b;
	double c;
	double d;
	double e;
	double f;

	/**
	 * Project using the following reference points:
	 *   (0,base_offset) -> point on projection.seed at t0
	 *   (1,base_offset) -> point on projection.seed at t1
	 * For other points:
	 *   (x,base_offset) -> point on projection.seed at t = t0 + x/(t1-t0)
	 *   (x,y)   -> point at a distance of base_offset-y from (x,base_offset)
	 */
	public Projection (Projection? parent, Seed seed, double t0, double t1, Flip flip) {
		this.parent = parent;
		this.seed = seed;
		// Identity
		a = 1; d = 1;
		// Rotate and scale about (0.5, 0.5), i.e., the centre of the segment
		//   translate -0.5, -0.5
		e = -0.5; f = -0.5;
		//   scale according to seed.tweak
		add (seed.tweak.scale_x, 0, 0, seed.tweak.scale_y);
		//   rotate according to seed.tweak
		var angle = seed.tweak.rotate * Math.PI / 180;
		double sin, cos;
		Math.sincos (angle, out sin, out cos);
		add (cos, sin, -sin, cos);
		//   translate according to seed.tweak
		// Translate 0.5, 0.5 to undo the very first translation
		e += seed.tweak.move_x + 0.5;
		f += seed.tweak.move_y + 0.5;
		// Flip
		if (flip.x ()) {    // x = 1 - x
			add (-1, 0, 0, 1);
			e += 1;
		}
		if (flip.y ()) {    // y = 1 - y
			add (1, 0, 0, -1);
			f += 1;
		}
		// Project onto the segment
		//  x = t0 + x * (t1 - t0)
		//  y = seed.base_offset - y
		add (t1 - t0, 0, 0, -1);
		e += t0;
		f += seed.base_offset;
	}

	public Projection.copy (Projection? parent, Projection source) {
		this.parent = parent;
		seed = source.seed;
		a = source.a;
		b = source.b;
		c = source.c;
		d = source.d;
		e = source.e;
		f = source.f;
	}

	/**
	 * Return p * abcedf * seed
	 */
	public static Point map (Projection? pr, Point p) {
		while (pr != null) {
		// Apply abcdef
			var x = pr.a * p.x  +  pr.c * p.y  +  pr.e;
			var y = pr.b * p.x  +  pr.d * p.y  +  pr.f;
			// Project the segment itself (possibly non-linear)
			p = pr.seed.distant_point (x, y);
			pr = pr.parent;
		}
		return p;
	}

	/**
	 * abcdef = abcdef * pqrs00
	 */
	void add (double p, double q, double r, double s) {
		var ta = a * p  +  b * r;
		var tb = a * q  +  b * s;
		var tc = c * p  +  d * r;
		var td = c * q  +  d * s;
		var te = e * p  +  f * r;
		var tf = e * q  +  f * s;
		a = ta; b = tb; c = tc; d = td; e = te; f = tf;
	}
}

public struct RGBA {
	public double r;
	public double g;
	public double b;
	public double a;

	public RGBA (double r, double g, double b, double a) {
		this.r = r;
		this.g = g;
		this.b = b;
		this.a = a;
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
